#include <stdio.h>
#include <stdlib.h> 
#include <string.h>
#include <assert.h>
#include "alphabet.h"
#include "utils.h"
#include "defs.h"

#define ALPHABET_LENGTH 256

#ifdef TOTO
int case_sensitive;
int skip_eor;
int line_mode;
#endif

typedef struct 
{
  letter_type type;
  char *equiv;
  int code;
} letter_entry;

static letter_entry alphabet[ALPHABET_LENGTH];
static int complement_table[MAXCHARS];

#ifndef DNA
extern int case_sensitive;
#endif
extern int skip_eor;
extern int line_mode;
char *separators;

#ifdef DNA
#define ASC2INT(c) ((unsigned char)(UPPER(c)))
#else
#define ASC2INT(c) ((unsigned char)(case_sensitive ? (c) : (UPPER(c))))
#endif

#ifndef ASC2INT
static unsigned int asc2int(const unsigned char c)
{
#ifdef DNA
  return((unsigned int)UPPER(c));
#else
  if(case_sensitive)
    {
      return((unsigned int)c);     
    }
  else
  {
    return((unsigned int)UPPER(c));
  }
#endif
}
#endif

static void add_direct_letter(char c,int code)
{
  int n;

#ifdef ASC2INT
  n=ASC2INT(c);
#else
  n=asc2int(c);
#endif

  alphabet[n].type=DIRECT;
  alphabet[n].code=code;
}

static void add_indirect_letter(char c,const char *s)
{
  int n;
  char *dup;

#ifdef ASC2INT
  n=ASC2INT(c);
#else
  n=asc2int(c);
#endif

  dup=strdup(s);
  assert(dup != NULL);

  alphabet[n].type=INDIRECT;
  alphabet[n].equiv=dup;
}

static void add_skip_letter(char c)
{
#ifdef ASC2INT
  alphabet[ASC2INT(c)].type=SKIP;
#else
  int n;

  n=asc2int(c);
  alphabet[n].type=SKIP;
#endif
}

static void add_peor_letter(char c)
{ 
#ifdef ASC2INT
  alphabet[ASC2INT(c)].type=PEOR;
#else
  int n;

  n=asc2int(c);
  alphabet[n].type=PEOR;
#endif
}

static void add_eor_letter(char c)
{
#ifdef ASC2INT
  alphabet[ASC2INT(c)].type=EOR;
#else
  int n;

  n=asc2int(c);
  alphabet[n].type=EOR;
#endif
}

static void reset_alphabet(void)
{
  int i;

  for(i=0;i<ALPHABET_LENGTH;i++)
    {
      alphabet[i].type=INVALID;
      alphabet[i].equiv=NULL;
      alphabet[i].code=-1;
    };
}

static void separator(void)
{
  char *p;
  char c;

  if(skip_eor)
    {
      if(separators == NULL)
	{
	  add_skip_letter('\n');
	}
      else
	{ 
	  p=separators;
	  while((c=*(p++)) != '\0')
	    {
	      add_eor_letter(c);
	    }
	}
    }
  
  if(line_mode)
    {
      if(separators == NULL)
	{
	  add_peor_letter('\r');
	  add_eor_letter('\n');
	}
      else
	{
	  p=separators;
	  while((c=*(p++)) != '\0')
	    {
	      add_eor_letter(c);
	    }
	}
    }
}

// ASCII-table
void init_alphabet(void)
{
  int i=0;
  
  alphabet[0].type=INVALID;
  alphabet[0].equiv=NULL;

  for(i=1;i<ALPHABET_LENGTH;i++)
    {
      alphabet[i].type=DIRECT;
      alphabet[i].code=i;
    }

  separator();
}

void init_iupac_code(void)
{
  reset_alphabet();

  //  case_sensitive=(0==1); // not case sensitive

  add_direct_letter('A',0); 
  add_direct_letter('C',1);
  add_direct_letter('G',2);
  add_direct_letter('T',3);
  add_direct_letter('U',3); // U is the same as T

  add_skip_letter(' ');

  separator();

  add_indirect_letter('R',"AG"); // [AG]
  add_indirect_letter('Y',"CT"); // [CT]
  add_indirect_letter('K',"GT"); // [GT]
  add_indirect_letter('M',"AC"); // [AC]
  add_indirect_letter('S',"CG"); // [CG]
  add_indirect_letter('W',"AT"); // [AT]
  add_indirect_letter('B',"CGT"); // [CGT]
  add_indirect_letter('D',"AGT"); // [AGT]
  add_indirect_letter('H',"ACT"); // [ACT]
  add_indirect_letter('V',"ACG"); // [ACG]
  add_indirect_letter('N',"ACGT"); // [ACGT]
}

letter_type get_letter_type(const char c)
{
#ifdef ASC2INT
  return alphabet[ASC2INT(c)].type;
#else
  return alphabet[asc2int(c)].type;
#endif
}

/*
int is_valid(char c)
{
#ifdef ASC2INT
  return (alphabet[ASC2INT(c)].type != INVALID);
#else
  return (alphabet[asc2int(c)].type != INVALID);
#endif
}

int is_letter(char c)
{
  letter_type t;

#ifdef ASC2INT
  t=alphabet[ASC2INT(c)].type;
#else
  t=alphabet[asc2int(c)].type;
#endif

  return ((t == DIRECT) || (t == INDIRECT));
}

int is_skip(char c)
{
#ifdef ASC2INT
  return (alphabet[ASC2INT(c)].type == SKIP);
#else
  return (alphabet[asc2int(c)].type == SKIP);
#endif
}

int is_peor(char c)
{
#ifdef ASC2INT
  return (alphabet[ASC2INT(c)].type == PEOR);
#else
  return (alphabet[asc2int(c)].type == PEOR);
#endif
}

int is_eor(char c)
{
#ifdef ASC2INT
  return (alphabet[ASC2INT(c)].type == EOR);
#else
  return (alphabet[asc2int(c)].type == EOR);
#endif
}

int is_indirect(char c)
{
#ifdef ASC2INT
  return (alphabet[ASC2INT(c)].type == INDIRECT);
#else
  return (alphabet[asc2int(c)].type == INDIRECT);
#endif
}

int is_direct(char c)
{
#ifdef ASC2INT
  return (alphabet[ASC2INT(c)].type == DIRECT);
#else
  return (alphabet[asc2int(c)].type == DIRECT);
#endif
}
*/

inline char *get_alias(const char c)
{
  // assert(is_indirect(c));

#ifdef ASC2INT
  return alphabet[ASC2INT(c)].equiv;
#else
  return alphabet[asc2int(c)].equiv;
#endif
}

inline int get_code(const char c)
{
  // assert(is_direct(c));

#ifdef ASC2INT
  return alphabet[ASC2INT(c)].code;
#else
  return alphabet[asc2int(c)].code;
#endif
}

char get_char(int code)
{
  char c;
  int i;

  for(i=0;i<256;i++)
    {
      c=(char)i;
      switch(get_letter_type(c))
	{
	case DIRECT:
	  if(get_code(c) == code)
	    {
	      return c;
	    }
	default:
	  break;
	}
    }
  return '\0';
}

char *get_simplification(const char *set)
{
  int i;
  char c;
  char res;
  char *p;
  int size=0;
  static int max_size=DFLT_SIZE;

  p=(char *)malloc(max_size*sizeof(char));
  assert(p != NULL);

  for(i=0;i<MAXCHARS;i++)
    {
      complement_table[i]=(0==1);
    }
  
  while((c=*(set++)) != '\0')
    {
      switch(c)
	{
	case BACKSLASH:
	  c=*(set++);
	  assert(c != '\0');
	  break;
	}

      switch(get_letter_type(c))
	{
	case DIRECT:
	  complement_table[get_code(c)]=(0==0);
	  break;
	case INDIRECT:
	  {
	    char *q;
	    char car;
	    q=get_alias(c);
	    while((car=*(q++)) != '\0')
	      {
		complement_table[get_code(car)]=(0==0);
	      }
	  }
	  break;
	default:
	  break;
	}
    }

  for(i=0;i<MAXCHARS;i++)
    {
      if(complement_table[i])
	{
	  if((res=get_char(i)) != '\0')
	    {
	      switch(res)
		{
		case DONTCARE:
		case PIPE:
		case BACKSLASH:
		case LEFT_SQ_BRACKET:
		case RIGHT_SQ_BRACKET:
		case LEFT_PAREN:
		case RIGHT_PAREN:
		case CIRCUMFLEX:
		case QUESTION:
		  ADD_CHAR_GROWING_BUFFER(BACKSLASH,p,size,max_size);
		}
	      ADD_CHAR_GROWING_BUFFER(res,p,size,max_size);
	    }
	}
    }
  p[size]='\0';
  return p;
}

char *get_complement(const char *set)
{
  int i;
  char c;
  char res;
  char *p;
  int size=0;
  static int max_size=DFLT_SIZE;

  p=(char *)malloc(max_size*sizeof(char));
  assert(p != NULL);

  for(i=0;i<MAXCHARS;i++)
    {
      complement_table[i]=(0==0);
    }
  
  while((c=*(set++)) != '\0')
    {
      switch(c)
	{
	case BACKSLASH:
	  c=*(set++);
	  assert(c != '\0');
	  break;
	}

      switch(get_letter_type(c))
	{
	case DIRECT:
	  complement_table[get_code(c)]=(0==1);
	  break;
	case INDIRECT:
	  {
	    char *q;
	    char car;
	    q=get_alias(c);
	    while((car=*(q++)) != '\0')
	      {
		complement_table[get_code(car)]=(0==1);
	      }
	  }
	  break;
	default:
	  break;
	}
    }

  for(i=0;i<MAXCHARS;i++)
    {
      if(complement_table[i])
	{
	  if((res=get_char(i)) != '\0')
	    {
	      switch(res)
		{
		case DONTCARE:
		case PIPE:
		case BACKSLASH:
		case LEFT_SQ_BRACKET:
		case RIGHT_SQ_BRACKET:
		case LEFT_PAREN:
		case RIGHT_PAREN:
		case CIRCUMFLEX:
		case QUESTION:
		  ADD_CHAR_GROWING_BUFFER(BACKSLASH,p,size,max_size);
		}
	      ADD_CHAR_GROWING_BUFFER(res,p,size,max_size);
	    }
	}
    }
  p[size]='\0';
  return p;
}

char *get_char_complement(const char c)
{
  static char tab[3];
  tab[0]=BACKSLASH;
  tab[1]=c;
  tab[2]='\0';
  return get_complement(tab);
}

static void print_alphabet(void)
{
  char c;
  int i;
  
  for(i=0;i<256;i++)
    {
      c=(char)i;
      switch(get_letter_type(c))
	{
	case DIRECT:
	  fprintf(stdout,"%c",c);
	default:
	  break;
	}
    }
  fprintf(stdout,"\n");
}

#ifdef TOTO
int main(void)
{
  char c;
  char *res;
  int i=0;

#ifdef DNA
  init_iupac_code();
#else
  init_alphabet();
#endif

  res=get_complement("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789");
  fprintf(stdout,"%d %s\n",strlen(res),res);
  res=get_complement(res);
  fprintf(stdout,"%d %s\n",strlen(res),res);
  res=get_char_complement('y');
  fprintf(stdout,"%d %s\n",strlen(res),res);

  return 0;
}
#endif
