#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include "dawg.h"
#include "alphabet.h"

#ifdef DAWG_SIZE
static int dawg_size=0;
static int max_dawg_size=0;
#endif

#ifdef DAWG_PRINT
static int num_print=0;
#endif

static int number=0;

extern node *source;
extern node *currentnode, *closestterminal;
extern int length;

static node *SPLIT(node *originnode,int a);
static void SET_SUF_POINTER(node *newchild,node *father);
static void DELETE_SUF_POINTER(node *oldchild,node *father);
static node *DELETE_LETTER(node *activenode);
static void MERGE(node *targetnode,node *originnode,int deletedletter);
static void DELETE_NODE(node *nod);

static kwlist *trapped_kwl=NULL;

#ifndef LET_TO_NUM
int LET_TO_NUM(char c)
{
  return get_code(c);
}
#endif

node *APPEND_LETTER(node *activenode,int letter)
{
  node *newactivenode,*varnode;
  
  assert((letter >= 0) && (letter <MAXCHARS));

  if ((newactivenode=activenode->out[letter]) != UNDEFINED) {
    if (activenode->type[letter] == PRIMARY) return(newactivenode);
    else return(SPLIT(activenode,letter));
  }
  else {
    newactivenode = CREATE_NODE();
    activenode->out[letter] = newactivenode; activenode->type[letter] = PRIMARY;
    newactivenode->origin = activenode;
    newactivenode->last_letter = letter;
    newactivenode->depth = activenode->depth + 1;
    varnode = activenode->suf_pointer;
    while (varnode != UNDEFINED && varnode->out[letter]==UNDEFINED) {
      varnode->out[letter] = newactivenode; varnode->type[letter] = SECONDARY;
      varnode = varnode->suf_pointer;
    }
    SET_SUF_POINTER(newactivenode,
		    (varnode == UNDEFINED) ? source
		    : ((varnode->type[letter] == PRIMARY) ?
		       varnode->out[letter] : SPLIT(varnode,letter)));
    return(newactivenode);
  }
}

static node *SPLIT(node *originnode,int a)
     /* modified version: the second parameter is a and not targetnode
	because SPLIT is always called SPLIT(x,x->out[a]) */
{
  node *targetnode,*newtargetnode,*n,*varnode;
  int appendedletter,c;
  
  targetnode = originnode->out[a];
  newtargetnode = CREATE_NODE();
  appendedletter = targetnode->last_letter;
  originnode->out[a] = newtargetnode; originnode->type[a] = PRIMARY;
  newtargetnode->origin = originnode;
  newtargetnode->last_letter = appendedletter;
  newtargetnode->depth = originnode->depth + 1;
  /* additional instructions */
  if (currentnode == targetnode) 
    if (length<=newtargetnode->depth) currentnode=newtargetnode; 
  /* end of additional instructions */
  for (c=0;c<MAXCHARS;c++)
    if ((n=targetnode->out[c]) != UNDEFINED) {
      newtargetnode->out[c] = n; newtargetnode->type[c] = SECONDARY;
    }
  SET_SUF_POINTER(newtargetnode,targetnode->suf_pointer);
  DELETE_SUF_POINTER(targetnode,targetnode->suf_pointer);
  SET_SUF_POINTER(targetnode,newtargetnode);
  /* after this: */
  /* targetnode has the suffix pointer to newtargetnode, */
  /* targetnode is the only child of newtargetnode, */
  varnode = originnode->suf_pointer;
  while (varnode != UNDEFINED && varnode->type[appendedletter]==SECONDARY) {
    varnode->out[appendedletter] = newtargetnode;
    varnode = varnode->suf_pointer;
  }
  return(newtargetnode);
}

static void SET_SUF_POINTER(node *newchild,node *father)
     /* it is supposed that newchild does not have a suffix pointer, */
     /* that is, does not belong to any child list. child is inserted */
     /* in the father's list at the leftmost position */
{
  node *current_leftmost_child;
  if (father->child !=UNDEFINED)
    {
      /* unnecessary test */
      if (father->number_of_children == 0 ) printf("problem5...\n");
      /* end of unnecessary test */
      current_leftmost_child=father->child;
      newchild->right_neighbour = current_leftmost_child;
      current_leftmost_child->left_neighbour = newchild;
    }
  newchild->suf_pointer=father;
  father->child=newchild;
  father->number_of_children++;
}

static void DELETE_SUF_POINTER(node *oldchild,node *father)
{
  if (oldchild->left_neighbour==UNDEFINED)
    if (oldchild->right_neighbour != UNDEFINED)
      /* oldchild is not the only child of father but is the leftmost one */
      {
	oldchild->right_neighbour->left_neighbour = UNDEFINED;
	father->child = oldchild->right_neighbour;
	oldchild->right_neighbour = UNDEFINED;
      }
    else
      /* oldchild is the only child of father */
      {
	father->child = UNDEFINED; 
      } 
  else
    if (oldchild->right_neighbour != UNDEFINED)
      /* oldchild has both right and left neighbour */
      {
	oldchild->right_neighbour->left_neighbour = oldchild->left_neighbour;
	oldchild->left_neighbour->right_neighbour = oldchild->right_neighbour;
	oldchild->left_neighbour = oldchild->right_neighbour = UNDEFINED;
      }
    else
      /* oldchild is not the only child of father but is the rightmost one */
      {
	oldchild->left_neighbour->right_neighbour = UNDEFINED;
	oldchild->left_neighbour = UNDEFINED;
      }
  oldchild->suf_pointer=UNDEFINED;
  father->number_of_children--;
}

node *LOAD_KEYWORD(char v[])
{
  int i; 
  node *activenode;
  
  activenode=source;
  for (i=0;i<strlen(v);i++) {
    activenode = APPEND_LETTER(activenode,LET_TO_NUM(v[i]));
    activenode->prefix_degree++;
  }
  return activenode;
}

node *CREATE_NODE(void)
{
  node *newnode;
  int i;

#ifdef DAWG_SIZE
  // calcul taille du dawg
  dawg_size++;
  max_dawg_size=(max_dawg_size>dawg_size)?(max_dawg_size):(dawg_size);
#endif

  newnode = (node*) malloc(sizeof(node)); 
  assert(newnode != NULL);
  for(i=0;i<MAXCHARS;i++)
    {
      newnode->out[i]=UNDEFINED;
    }
  newnode->suf_pointer=newnode->origin=UNDEFINED;
  newnode->child=newnode->left_neighbour=newnode->right_neighbour=UNDEFINED;
  newnode->terminal=NULL;
  newnode->number_of_children=newnode->prefix_degree=0;
  newnode->depth=0;
  newnode->visited=NOT_YET;
  newnode->last_letter=-1;
  newnode->number=++number;
  return newnode;
}

static node *DELETE_LETTER(node *activenode)
{
  node *newactivenode,*varnode,*suffixnode; 
  int deletedletter;

  newactivenode = activenode->origin;
  deletedletter = activenode->last_letter;
  if (activenode->prefix_degree>0 || 
      activenode->number_of_children > 1) return(newactivenode);
  else if (activenode->number_of_children==1) {
    MERGE(activenode,newactivenode,deletedletter);
    return(newactivenode);
  }
  else {
    newactivenode->out[deletedletter]=UNDEFINED;
    /* additional instructions */
    if (currentnode==activenode) { 
      currentnode=activenode->suf_pointer; 
      length = currentnode->depth; 
    }
    if (closestterminal==activenode){
      closestterminal=CLOSEST_TERMINAL(closestterminal);
    }
    
    /* end of additional instructions */
    varnode=newactivenode->suf_pointer;
    while (varnode!=UNDEFINED && varnode->type[deletedletter]==SECONDARY) {
      varnode->out[deletedletter]=UNDEFINED;
      varnode = varnode->suf_pointer;
    }
    DELETE_SUF_POINTER(activenode,activenode->suf_pointer);
    DELETE_NODE(activenode); 
    if (varnode != UNDEFINED) {
      suffixnode = varnode->out[deletedletter];
      if (suffixnode->prefix_degree==0 && suffixnode->number_of_children==1)
	MERGE(suffixnode,varnode,deletedletter);
    }
    return(newactivenode);
  }
}

static void MERGE(node *targetnode,node *originnode,int deletedletter)
/* targetnode must have exactly one child */
{
  node *newtargetnode,*varnode;
  
  newtargetnode = targetnode->child;
  /* additional instructions */
  if (currentnode == targetnode) currentnode=newtargetnode;
  if ((closestterminal!=NULL)&&(closestterminal==targetnode)){
    closestterminal=CLOSEST_TERMINAL(closestterminal);
  }
  
  /* end of additional instructions */
  originnode->out[deletedletter] = newtargetnode;
  originnode->type[deletedletter] = SECONDARY;
  varnode = originnode->suf_pointer;
  while (varnode!=UNDEFINED && varnode->type[deletedletter]==SECONDARY) {
    varnode->out[deletedletter] = newtargetnode;
    varnode = varnode->suf_pointer;
  }
  DELETE_SUF_POINTER(newtargetnode,targetnode);
  SET_SUF_POINTER(newtargetnode,targetnode->suf_pointer);
  DELETE_SUF_POINTER(targetnode,targetnode->suf_pointer);
  DELETE_NODE(targetnode);
}

static void DELETE_NODE(node *nod)
     /* it is supposed that no edges point to nod and that */
     /* the suffix pointer of nod has been deleted */
{
#ifdef DAWG_SIZE
  // calcul taille dawg
  dawg_size--;
#endif

  free(nod);
}

void UNLOAD_KEYWORD(node *terminalnode)
{
  node *activenode;
  
  activenode = terminalnode;
  while (activenode != source) {
    activenode->prefix_degree--;
    activenode = DELETE_LETTER(activenode);
  }
}

node *CLOSEST_TERMINAL(node *nod)   
{
  node *aux=nod;
  while ((aux=aux->suf_pointer) != NULL)  
    if (aux->terminal != NULL) return(aux);
  return UNDEFINED;
}

kwlist *new_kwnode(void)
{
  kwlist *kwl;

  kwl=malloc(sizeof(kwlist));
  assert(kwl != NULL);
  return kwl;
}

void trap_kwlist(kwlist *kwl)
{
  trapped_kwl=kwl;
}

kwlist *untrap_kwlist(void)
{
  kwlist *res;
  res=trapped_kwl;
  trapped_kwl=NULL;
  return res;
}

void clear_kwlist(node *terminal_node,void *word)
{
  kwlist *old_kwl,*kwl;

  assert(word != NULL);
  old_kwl=NULL;
  kwl=terminal_node->terminal;

  // fprintf(stderr,"clear_kwlist %p\n",word);
  while(kwl != NULL)
    {
      // fprintf(stderr,"%p %p %p\n",old_kwl,kwl,kwl->word);
      if(kwl->word == word)
	{
	  break;
	}
      old_kwl=kwl;
      kwl=kwl->next;
    }
  // fprintf(stderr,"bug\n");
  assert(kwl != NULL);

  if(kwl == trapped_kwl)
    {
      // fprintf(stderr,"trap\n");
      trapped_kwl=kwl->next;
    }

  if(old_kwl == NULL)
    {
      terminal_node->terminal=kwl->next;
      kwl->next=NULL;
      free(kwl);
    }
  else
    {
      // fprintf(stderr,"old: %p\nact: %p\ntrapped: %p\nter: %p\n",old_kwl,kwl,trapped_kwl,terminal_node->terminal);
      // {
      // kwlist *tmp;
      // tmp=terminal_node->terminal;
      // while(tmp != NULL)
      // {
      // fprintf(stderr,"%p ",tmp);
      // tmp=tmp->next;
      // }
      // fprintf(stderr,"\n");
      // }
      old_kwl->next=kwl->next;
      kwl->next=NULL;
      free(kwl);
    }
}

void delete_kwlist(kwlist *kwl)
{
  kwlist *tmp;

  while(kwl != NULL)
    {
      tmp=kwl->next;
      kwl->next=NULL;
      free(kwl);
      kwl=tmp;
    }
}

#ifdef DAWG_PRINT
static void PRINT_DAWG(node *source)
{
  int i;

  if(source == NULL)
    {
      return;
    }

  for(i=0;i<MAXCHARS;i++)
    {
      if(source->out[i] != NULL)
	{
	  if(source->type[i] == PRIMARY)
	    {
	      fprintf(stdout,"C%d-Pri ",num_print);
	      fprintf(stdout,"%d ",i);
	      fprintf(stdout,"%d %d\n",source->number,source->out[i]->number);
	      PRINT_DAWG(source->out[i]);
	    }
	  else
	    {
	      fprintf(stdout,"D%d-Sec ",num_print);
	      fprintf(stdout,"%d ",i);
	      fprintf(stdout,"%d %d\n",source->number,source->out[i]->number);
	    }
	}
    }
  if(source->suf_pointer != UNDEFINED)
    {
      fprintf(stdout,"E%d-Suf ",num_print);
      fprintf(stdout,"%d %d\n",source->number,source->suf_pointer->number);
    }
  if(source->terminal != NULL)
    {
      fprintf(stdout,"B%d-Ter ",num_print);
      fprintf(stdout,"%d %d\n",source->depth,source->number);
    }
  else
    {
      fprintf(stdout,"A%d-Sta ",num_print);
      fprintf(stdout,"%d %d\n",source->depth,source->number);
    }
}

void NEW_PRINT_DAWG(node *source)
{
  num_print++;
  PRINT_DAWG(source);
}
#endif

void FREE_DAWG(node *source)
{
  int i;

#ifdef DAWG_PRINT
  if(number != 0)
    {
      NEW_PRINT_DAWG(source);
      number=0;
    }
#endif

#ifdef DAWG_SIZE
  // calcul taille dawg
  if(dawg_size != 0)
    {
      fprintf(stdout,"Max DAWG Size: %d\n",max_dawg_size);
      dawg_size=0;
    }
#endif

  if(source != NULL)
    {
      
      for(i=0;i<MAXCHARS;i++)
	{
	  if((source->out[i] != NULL) && (source->type[i] == PRIMARY))
	    {
	      FREE_DAWG(source->out[i]);
	      source->out[i]=NULL;
	    }
	}
      
      delete_kwlist(source->terminal);
      source->terminal=NULL;
      free(source);
    }              
}
