#ifndef __DAWG_H
#define __DAWG_H

#include "defs.h"
#include "alphabet.h"

#define PRIMARY 'P'
#define SECONDARY 'S'

#define ALREADY 'A'
#define NOT_YET 'B'

typedef struct NODE {
  struct NODE *out[MAXCHARS];  
  char type[MAXCHARS]; 
  struct NODE *suf_pointer;
  int depth;  
  struct KWLIST *terminal;     /* list "terminal" */
  struct NODE *origin;
  int last_letter;
  int number_of_children;       /* we can do without this attribut */
  struct NODE *child;           /* references to the "leftmost" child */
  struct NODE *left_neighbour;  /* "left neighbour" in the children list */
  struct NODE *right_neighbour; /* "right neighbour" in the children list */
  int prefix_degree;
  char visited;              /* auxiliary field; for printing */
  int number;                /* auxiliary field; for numbering */
} node;

typedef struct PAIR {         /* pair <state,length> */
  node *nod;
  int len;
} pair;

typedef struct KWLIST {     /* element type of the list "terminal" */
  void *word; /* pointer to keyword for which the node is terminal */
  void *frag_cell; /* fragment cell of the keyword */
  void *pat_frag; /*  pattern fragment that contains the fragment */
  void *pattern; /* pattern of the fragment */
  struct KWLIST *next;  /* next element in the list */
} kwlist;

node *APPEND_LETTER(node *activenode,int letter);
node *CREATE_NODE(void);
node *LOAD_KEYWORD(char v[]);
void UNLOAD_KEYWORD(node *terminalnode);
node *CLOSEST_TERMINAL(node *nod);

kwlist *new_kwnode(void);
void trap_kwlist(kwlist *kwl);
kwlist *untrap_kwlist(void);
void clear_kwlist(node *terminalnode,void *word);
#ifdef DAWG_PRINT
void NEW_PRINT_DAWG(node *source);
#endif
void FREE_DAWG(node *source);

#define LET_TO_NUM(c) get_code(c)
// int LET_TO_NUM(char c);

#endif // __DAWG_H
