#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h> 
#include "keyword.h"
#include "alphabet.h"
#include "dawg.h"
#include "defs.h"

// the string parent is duplicated by strdup
keyword_set *new_keyword_set(const char *parent)
{
  keyword_set *new_set;
  char *tmp;

  new_set=(keyword_set *)malloc(sizeof(keyword_set));
  assert(new_set != NULL);

  tmp=strdup(parent);
  assert(tmp != NULL);

  new_set->parent=tmp;
  new_set->type=COLLAPSED;
  new_set->index=0;
  new_set->length=0;
  new_set->children=empty_keyword_list();
  new_set->positions=empty_int_list();

  return new_set;
}

static keyword *new_keyword(int nbchars,const char *init)
{
  keyword *w;
  char *p;
  int len=0;

  w=(keyword *)malloc(sizeof(keyword));
  assert(w != NULL);

  p=(char *)malloc((nbchars+1)*sizeof(char));
  assert(p != NULL);

  w->keyword=p;
  w->set=NULL;
  w->portnode=NULL;
  w->terminal_node=NULL;
  
  if(init != NULL)
    {
      strncpy(p,init,nbchars);
      len=strlen(init);
      len=(len>nbchars)?nbchars:len;
      len++;
    }
  *(p+len)='\0';
  
  //fprintf(stdout,"%d %s\n",strlen(w->keyword),w->keyword);
  
  return w;
}

static void delete_keyword(keyword *w)
{
  assert(w != NULL);

  if(w->keyword != NULL)
    {  
      free(w->keyword);
      w->keyword=NULL;
    }
  w->set=NULL;
  w->portnode=NULL;
  w->terminal_node=NULL;

  free(w);
}

/*******************************************************/
/* copy_keyword_list copy a **non-empty** keyword_list */
/* the source list must be in a set (not NULL)         */
/* the length of the strings is based on the length    */
/* of the parent of the keyword_set                    */
/* Note: the copy is reversed                          */
/*******************************************************/
static keyword_list copy_keyword_list(keyword_list l,int nbchars)
{
  // int nbchars;
  keyword *w,*w2;
  keyword_list l2;

  l2=empty_keyword_list();

  // nbchars=strlen(((keyword *)keyword_list_car(l))->set->parent);;

  // fprintf(stderr,"copy_keyword_list: %d\n",nbchars);

  while(!is_empty_keyword_list(l))
    {
      w=(keyword *)keyword_list_car(l);
      l=keyword_list_cdr(l);

      w2=new_keyword(nbchars,w->keyword);
      w2->set=w->set;
      w2->portnode=w->portnode;
      w2->terminal_node=w->terminal_node; // should be NULL

      l2=keyword_list_cons((struct keyword *)w2,l2);
    }

  return l2;
}

static void print_keyword_list(keyword_list l)
{
  keyword *w;
  fprintf(stdout,"keyword list:\n");

  while(!is_empty_keyword_list(l))
    {
      w=(keyword *)keyword_list_car(l);
      l=keyword_list_cdr(l);

      fprintf(stdout,"%s %p %p %p\n",w->keyword,w->set,w->portnode,w->terminal_node);
    } 
}

void print_keyword_set(keyword_set *s)
{
  fprintf(stdout,"%s\n",s->parent);
  fprintf(stdout,"index: %d\n",s->index);
  fprintf(stdout,"length: %d\n",s->length);
  print_keyword_list(s->children);
}

/* it is assumed that s is enough large */
static void add_char_to_string(char c,char *s,int end_pos)
{
  s=s+end_pos;
  *(s++)=c;
  *s='\0';
  /*  
      while(*(s++) != '\0');
      *s='\0'; // new end of string
      *(--s)=c;
      */
}

static keyword_list add_char_to_keyword_list(char c,keyword_list l,int end_pos)
{
  keyword *w;
  keyword_list tmp;
  tmp=l;

  assert(get_letter_type(c) == DIRECT);

  while(!is_empty_keyword_list(l))
    {
      w=(keyword *)keyword_list_car(l);
      l=keyword_list_cdr(l);
      
      add_char_to_string(c,w->keyword,end_pos);
    }

  return tmp;
}

static keyword_list add_string_to_keyword_list(const char *s,keyword_list l,int end_pos,int nbchars)
{
  keyword_list tmp,res;
  char c;
  
  if(*s == '\0')
    {
      return l;
    }

  res=empty_keyword_list();

  while((c=*(s++)) != '\0')
    {
      if(c == BACKSLASH)
	{
	  c=(*s++);
	  assert(c != '\0');
	}

      if(*s == '\0')
	{
	  tmp=l;
	}
      else
	{
	  tmp=copy_keyword_list(l,nbchars);
	}
      add_char_to_keyword_list(c,tmp,end_pos);
      res=keyword_list_append(tmp,res);
    }

  return res;
}

/* at the end of the procedure, '*len' is the exact length of the generated strings */
/* w is an empty keyword */
static keyword_list generate_set(keyword *w,char *p,int *len)
{
  char c;
  char *alias;
  keyword_list l;
  int length=0;
  int max_nbchars;

  max_nbchars=strlen(p);

  l=keyword_list_cons((struct keyword *)w,empty_keyword_list());
  while((c=*(p++)) != '\0')
    {
      alias=NULL;
      switch(c)
	{
	case LEFT_SQ_BRACKET:
	  {
	    char *alias;
	    char tmp;

	    alias=p;
	    while((tmp=*(p++)) != '\0')
	      {
		if(tmp == RIGHT_SQ_BRACKET)
		  {
		    break;
		  }

		if(tmp == BACKSLASH)
		  {
		    tmp=*(p++);
		    assert(tmp != '\0');
		  }
	      }

	    assert(tmp == RIGHT_SQ_BRACKET);

	    *(--p)='\0';
	    if(*alias == CIRCUMFLEX)
	      {
		char *compl;
		compl=get_complement(alias+1);
		if(*compl != '\0')
		  {
		    // fprintf(stderr,"%s\n",compl);
		    l=add_string_to_keyword_list(compl,l,length++,max_nbchars);
		  }
		free(compl);
	      }
	    else
	      {
		char *simpl;
		simpl=get_simplification(alias);
		if(*simpl != '\0')
		  {
		    // fprintf(stderr,"%s\n",simpl);
		    l=add_string_to_keyword_list(simpl,l,length++,max_nbchars);
		  }
		free(simpl);
	      }
	    *(p++)=RIGHT_SQ_BRACKET;
	  }
	  break;
	case BACKSLASH:
	  c=(*p++);
	  assert(c != '\0');
	  /* act as if it was an usual character */
	default:
	  if(get_letter_type(c) == DIRECT)
	    {
	      l=add_char_to_keyword_list(c,l,length++);
	    }
	  else
	    if(get_letter_type(c) == INDIRECT)
	    {
	      l=add_string_to_keyword_list(get_alias(c),l,length++,max_nbchars);
	    }
	  break;
	}
    }
  *len=length;
  return l;
}

/* The following function generates all the words at a distance 1 (substitution) from the parent of the keyword set s */
/* note that the keyword set must have been expanded before, in particular s->length must be a upper bound of the     */
/* size of the generated strings */
static keyword_list generate_one_error(keyword_set *s,node *source)
{
  char c;
  char *start;
  char *p,*q;
  char *buffer;
  char *compl;
  int size;
  static int max_size=DFLT_SIZE;
  int max_nbchars;
  int valid;
  int length;
  keyword *w;
  keyword_list res;

  //  max_nbchars=strlen(s->parent);
  max_nbchars=s->length;

  buffer=(char *)malloc(max_size*sizeof(char));
  assert(buffer != NULL);

  res=empty_keyword_list();

  for(p=s->parent;(*(p++)) == QUESTION;);
  p--;
  start=p;

  while((c=*(p++)) != '\0')
    {
      valid=(0==0);
      size=0;

      for(q=start;q<p-1;)
	{
	  ADD_CHAR_GROWING_BUFFER(*(q++),buffer,size,max_size);
	}

      switch(c)
	{
	case LEFT_SQ_BRACKET:
	  {
	    char *alias;
	    char tmp;
	    
	    alias=p;
	    while((tmp=*(p++)) != '\0')
	      {
		if(tmp == RIGHT_SQ_BRACKET)
		  {
		    break;
		  }
		
		if(tmp == BACKSLASH)
		  {
		    tmp=*(p++);
		    assert(tmp != '\0');
		  }
	      }
	    
	    assert(tmp == RIGHT_SQ_BRACKET);
	    
	    *(--p)='\0';
	    
	    ADD_CHAR_GROWING_BUFFER(LEFT_SQ_BRACKET,buffer,size,max_size);
	    if(*alias == CIRCUMFLEX)
	      {
		alias++;
	      }
	    else
	      {
		ADD_CHAR_GROWING_BUFFER(CIRCUMFLEX,buffer,size,max_size);
	      }

	    compl=get_complement(alias);
	    if((*compl == '\0') || (*alias == '\0'))
	      {
		valid=(0==1);
	      }
	    free(compl);
	    
	    for(q=alias;(tmp=*(q++)) != '\0';)
	      {
		ADD_CHAR_GROWING_BUFFER(tmp,buffer,size,max_size);
	      }
	    ADD_CHAR_GROWING_BUFFER(RIGHT_SQ_BRACKET,buffer,size,max_size);

	    *(p++)=RIGHT_SQ_BRACKET;
	  }
	  break;
	case BACKSLASH:
	  c=*(p++);
	  assert(c != '\0');
	  ADD_CHAR_GROWING_BUFFER(LEFT_SQ_BRACKET,buffer,size,max_size);
	  ADD_CHAR_GROWING_BUFFER(CIRCUMFLEX,buffer,size,max_size);
	  ADD_CHAR_GROWING_BUFFER(BACKSLASH,buffer,size,max_size);

	  compl=get_char_complement(c);
	  if(*compl == '\0')
	    {
	      valid=(0==1);
	    }
	  free(compl);

	  ADD_CHAR_GROWING_BUFFER(c,buffer,size,max_size);
	  ADD_CHAR_GROWING_BUFFER(RIGHT_SQ_BRACKET,buffer,size,max_size);
	  break;
	default:
	  ADD_CHAR_GROWING_BUFFER(LEFT_SQ_BRACKET,buffer,size,max_size);
	  ADD_CHAR_GROWING_BUFFER(CIRCUMFLEX,buffer,size,max_size);

	  compl=get_char_complement(c);
	  if(*compl == '\0')
	    {
	      valid=(0==1);
	    }
	  free(compl);
	  
	  ADD_CHAR_GROWING_BUFFER(c,buffer,size,max_size);
	  ADD_CHAR_GROWING_BUFFER(RIGHT_SQ_BRACKET,buffer,size,max_size);
	  break;
	}
      
      for(q=p;(c=*(q++)) != '\0';)
	{
	  ADD_CHAR_GROWING_BUFFER(c,buffer,size,max_size);
	}

      buffer[size]='\0';
      if(valid)
	{
	  w=new_keyword(max_nbchars,NULL);
	  w->set=s;
	  w->portnode=source;
	  w->terminal_node=NULL;

	  //fprintf(stderr,"%s\n",buffer);
	  res=keyword_list_append(generate_set(w,buffer,&length),res);
	  //fprintf(stderr,"%d %d\n",s->length,length);
	  assert(s->length == length);
	}
    }
  free(buffer);

  return res;
}

void expand_keyword_set(keyword_set *s,node *source)
{
  char *p;
  int nb_errors;
  keyword *w;
 
  if(s->type == EXPANDED)
    {
      return;
    }

  // fprintf(stderr,"expand keyword set\n");
  
  for(p=s->parent,nb_errors=0;(*(p++)) == QUESTION;nb_errors++);
  p--;

  w=new_keyword(strlen(s->parent),NULL);
  w->set=s;
  w->portnode=source;
  w->terminal_node=NULL;

  s->children=generate_set(w,p,&(s->length));
  s->index=0;

  switch(nb_errors)
    {
    case 0:
      break;
    default:
      fprintf(stderr,"warning: grappe can handle at most 1 error by keyword\n");
    case 1:
      s->children=keyword_list_append(s->children,generate_one_error(s,source));
      break;
    }

#ifdef DEBUG
  fprintf(stdout,"%d\n",s->length);
  print_keyword_list(s->children);
#endif

  s->type=EXPANDED;
}

void collapse_keyword_set(keyword_set *s)
{
  keyword_list l;
  keyword *w;

  if(s->type == COLLAPSED)
    {
      return;
    }

  l=s->children;

  while(!is_empty_keyword_list(l))
    {
      w=(keyword *)keyword_list_car(l);
      delete_keyword(w);
      l=keyword_list_cdr(l);
    }

  delete_keyword_list(s->children);
  s->children=empty_keyword_list();

  s->type=COLLAPSED;
}

int is_loaded(keyword_set *s)
{
  if(s->type == COLLAPSED)
    {
      return (s->index != 0);
    }
  else
    {
      return (s->index == s->length);
    }
}

// the real prototype of the following function is:
// int load_keyword_set(keyword_set *s,int position,node *source,fragment_cell *frag_cell,pattern *pat,pattern_fragment *pat_frag)
// see keyword.h for details
int load_keyword_set(keyword_set *s,int position,node *source,void *frag_cell,void *pat,void *pat_frag)
{
  char patternletter;
  int index;
  keyword *w;
  keyword_list elt;
  node *newportnode;

  expand_keyword_set(s,source);
  elt=s->children;
  index=(s->index)++;
  
  while(!is_empty_keyword_list(elt))
    {
      w=(keyword *)keyword_list_car(elt);
      patternletter=w->keyword[index];
      // fprintf(stderr,"%p %s %c %p\n",w,w->keyword,patternletter,w->portnode);
      newportnode=APPEND_LETTER(w->portnode,LET_TO_NUM(patternletter));
      newportnode->prefix_degree++;
      if(s->index == s->length)
	{
	  kwlist *kl;

	  w->terminal_node=newportnode;

	  kl=new_kwnode();
	  kl->word=w;
	  kl->frag_cell=frag_cell;
	  kl->pattern=pat;
	  kl->pat_frag=pat_frag;
	  kl->next=newportnode->terminal;
	  newportnode->terminal=kl;
	}
      else
	{
	  w->portnode=newportnode;
	}
      elt=keyword_list_cdr(elt);
    }

  return (s->index == s->length);
}

void unload_keyword_set(keyword_set *s)
{
  keyword *w;
  keyword_list elt;

  elt=s->children;

  while(!is_empty_keyword_list(elt))
    {
      w=(keyword *)keyword_list_car(elt);

      clear_kwlist(w->terminal_node,w);
      UNLOAD_KEYWORD(w->terminal_node);

      elt=keyword_list_cdr(elt);
    }

  delete_int_list(s->positions);
  s->positions=empty_int_list();

  // collapse_keyword_set(s);
}

void reset_keyword_set(keyword_set *s,node *source)
{
  s->index=0;

  delete_int_list(s->positions);
  s->positions=empty_int_list();

  if(s->type == EXPANDED)
    {
      keyword_list l;
      keyword *w;
      
      l=s->children;

      while(!is_empty_keyword_list(l))
	{
	  w=(keyword *)keyword_list_car(l);
	  w->portnode=source;
	  w->terminal_node=NULL;
	  l=keyword_list_cdr(l);
	}
    }
}
