#ifndef __KEYWORD_H
#define __KEYWORD_H

#include "lists.h"
#include "dawg.h"

typedef enum {COLLAPSED,EXPANDED} keyword_type;

typedef struct
{
  char *parent;  // generic keyword (collapsed)
  keyword_type type; // indicate whether parent keyword has been expanded or not
  int index;   // indicate where we are in the loading of the keywords
  int length;  // length of the expanded keywords (note that it is constant all over the set)
  keyword_list children; // list of the expanded keywords
  int_list positions;
} keyword_set;

typedef struct
{
  char *keyword; // expanded keyword
  keyword_set *set; // pointer to the keyword set
  node *portnode;
  node *terminal_node;
} keyword;

keyword_set *new_keyword_set(const char *parent);
void expand_keyword_set(keyword_set *s,node *source);
void collapse_keyword_set(keyword_set *s);
void print_keyword_set(keyword_set *s);
int is_loaded(keyword_set *s);
// the real prototype of the following function is:
// int load_keyword_set(keyword_set *s,int position,node *source,fragment_cell *frag_cell,pattern *pat,pattern_fragment *pat_frag);
// for some reasons, I didn't manage to include "pattern.h" and "fragment.h" in order to declare missing types
int load_keyword_set(keyword_set *s,int position,node *source,void *frag_cell,void *pat,void *pat_frag);
void unload_keyword_set(keyword_set *s);
void reset_keyword_set(keyword_set *s,node *source);

#endif // __KEYWORD_H
