#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "defs.h"
#include "alphabet.h"
#include "errors.h"
#include "dawg.h"
#include "pattern.h"

node *source;
node *currentnode,*closestterminal;
int length;
int num_of_pats;

extern int line_mode;
extern int counter_mode;

extern int counter;

static int num_line;

pattern *p[MAXPAT];

static void init_match(int *position)
{
  int i;
  
  source=CREATE_NODE();
  currentnode=source;
  length=0;
  *position=0;

  for(i=0;i<num_of_pats;i++)
    {
      p[i]->underloading=(0==0);
      reset_pattern(p[i],source);
    }
}

static void UPDATE_CURRENT_NODE(int currentletter)
{
  while(currentnode->out[currentletter] == UNDEFINED)
    {
      if(currentnode == source)
	{
	  length=0;
	  return;
	}
      else
	{
	  currentnode=currentnode->suf_pointer;
	  length=currentnode->depth;
	}
    }
  currentnode=currentnode->out[currentletter];
  length++;
}

static void stage1(int position)
{
  int i;
#ifdef DAWG_PRINT
  int print_flag;
#endif

  for(i=0;i<num_of_pats;i++)
    {
#ifdef DAWG_PRINT
      print_flag=p[i]->underloading;
#endif
      load_pattern(p[i],position,source);
#ifdef DAWG_PRINT
      if(print_flag)
	{
	  NEW_PRINT_DAWG(source);
	}
#endif
    }
}

static int stage3(int position)
{
  kwlist *kwl;
  node *currentterminal;
  pattern *pat;
  fragment_cell *frag_cell;
  keyword *word;
  keyword_set *set;
  pattern_fragment *pat_frag;

  if((currentnode->terminal != NULL) && (currentnode->depth == length))
    {
      closestterminal=currentnode;
    }
  else
    {
      closestterminal=CLOSEST_TERMINAL(currentnode);
    }

  while(closestterminal != UNDEFINED)
    {
      currentterminal=closestterminal;
      closestterminal=CLOSEST_TERMINAL(closestterminal);

      kwl=currentterminal->terminal;
      while(kwl != NULL)
	{
	  word=kwl->word;
	  frag_cell=kwl->frag_cell;
	  pat=kwl->pattern;
	  pat_frag=kwl->pat_frag;
	  set=word->set;

	  /* a fragment is: v_0(p_0,q_0)v_1...(p_n-1,q_n-1)v_n */
	  /* here we have found v_k */
	  if(frag_cell->next == NULL)
	    {
	      // print_fragment(frag_cell);
	      // here, k=n
	      if(frag_cell->previous == NULL)
		{
		  // k=n=0
		  set->positions=int_list_insert(position,set->positions);

		  pat->underloading=(0==0);
		  pat_frag->end_position=position;
		  pat_frag->occurs=(0==0);

		  trap_kwlist(kwl);
		  unload_fragment(pat_frag->fragment);
		  kwl=untrap_kwlist();
		  // fprintf(stderr,"%p %p\n",kwl,currentterminal->terminal);
		  // BUG
		  // kwl=currentterminal->terminal;
		}
	      else
		{
		  // the pattern occurs iff v_n-1 occurs at a possible position
		  /* fprintf(stderr,"checking positions (last keyword)\n"); */
		  if(previous_keyword_occurs(frag_cell->previous,position-set->length))
		    {
		      /* fprintf(stderr,"linked fragment occurs\n"); */
		      set->positions=int_list_insert(position,set->positions);

		      pat->underloading=(0==0);
		      pat_frag->end_position=position;
		      pat_frag->occurs=(0==0);

		      trap_kwlist(kwl);
		      unload_fragment(pat_frag->fragment);
		      kwl=untrap_kwlist();
		      // fprintf(stderr,"%p %p\n",kwl,currentterminal->terminal);
		      // BUG
		      // kwl=currentterminal->terminal;
		    }
		  else
		    {
		      kwl=kwl->next;
		    }
		}

	      if(pattern_occurs(pat))
		{
		  counter++;
		  if(!counter_mode)
		    {
		      if(line_mode)
			{
			  fprintf(stdout,"line %d: ",num_line);
			}
		      print_pattern(pat);
		      fprintf(stdout," occurs at position ");
		      print_pattern_position(pat);
		      fprintf(stdout,"\n");
		    }
		  return (0==0);
		}
	    }
	  else
	    {
	      // here k != n
	      if(frag_cell->previous == NULL)
		{
		  // here k=0
		  // we insert the position in the list of positions
		  set->positions=int_list_insert(position,set->positions);
		  kwl=kwl->next;
		  /* fprintf(stderr,"first keyword\n"); */
		}
	      else
		{
		  // 0 < k < n
		  // we add the position only if we found a correct position of v_k-1
		  /* fprintf(stderr,"checking positions\n"); */
		  if(previous_keyword_occurs(frag_cell->previous,position-set->length))
		    {
		      /* fprintf(stderr,"constraint satisfied\n"); */
		      set->positions=int_list_insert(position,set->positions);
		    }
		  kwl=kwl->next;
		}
	    }
	}
    }
  return (0==1);
}

static int skip_file(FILE *fd)
{
  char c,next;

  if(!line_mode)
    {
      return (0==0);
    }

  while((c=fgetc(fd)) != EOF)
    {
      switch(get_letter_type(c))
	{
	case PEOR:
	  while(get_letter_type(next=fgetc(fd)) == PEOR);
	  if(get_letter_type(next) != EOR)
	    {
	      ungetc(next,fd);
	    }
	case EOR:
	  num_line++;
	  return (c == EOF);
	default:
	}
    }
  return (c == EOF);
}

int match(FILE *fd)
{
  int position;
  char currentletter,nextletter;
  int found;

  found=(0==1);

  num_line=1;
  init_match(&position);

  while((currentletter=fgetc(fd)) != EOF)
    {
      // fprintf(stderr,"%c",currentletter);
      switch(get_letter_type(currentletter))
	{
	case DIRECT:
	  position++;
	  break;
	case SKIP:
	  continue;
	case PEOR:
	  while(get_letter_type(nextletter=fgetc(fd)) == PEOR);
	  if(get_letter_type(nextletter) != EOR)
	    {
	      ungetc(nextletter,fd);
	    }
	  // do as if it was an End Of Record
	case EOR:
	  num_line++;
	  FREE_DAWG(source);
	  source=NULL;
	  init_match(&position);
	  continue;
	default:
	  fprintf(stderr,"error while parsing text (illegal char %u)\n",(unsigned char)currentletter);
	  // exit(ILLEGAL_CHAR);
	  return (0==1);
	}
      
      // stage 1
      stage1(position);
      
      // stage 2
      UPDATE_CURRENT_NODE(LET_TO_NUM(currentletter));

      // stage 3
      if(stage3(position))
	{
	  found=(0==0);
	  FREE_DAWG(source);
	  source=NULL;
	  if(!skip_file(fd))
	    {
	      init_match(&position);
	      continue;
	    }

	  return found;
	}
    }
  FREE_DAWG(source);
  source=NULL;
  return found;
}
