#include <assert.h>
#include <stdio.h>
#include "errors.h"
#include "commands.h"
#include "parse_pattern.h"
#include "alphabet.h"

#ifndef DNA
int case_sensitive=(0==0);
#endif
int skip_eor=(0==0);
int line_mode=(0==1);
int counter_mode=(0==1);
extern char *separators;

static char *progname=NULL;
static char *pattern=NULL;
static char *pattern_filename=NULL;

static void usage(void)
{
  fprintf(stderr,"grappe %s ",VERSION);
#ifdef DNA
  fprintf(stderr,"(compiled with --enable-DNA option)");
#endif
  fprintf(stderr,"\n");
  fprintf(stderr,"usage: %s ",progname);
  fprintf(stderr,"[-%c patternfile] ",PATTERN_FILE);
  fprintf(stderr,"[-%c] ",SKIP_EOR);
  fprintf(stderr,"[-%c] ",LINE_MODE);
  fprintf(stderr,"[-%c] ",COUNTER_MODE);
#ifndef DNA
  fprintf(stderr,"[-%c] ",CASE_SENSITIVE);
#endif
  fprintf(stderr,"[-%c separators] ",DEFINE_SEPARATORS);
  fprintf(stderr,"[-%c] pattern ",PATTERN);
  fprintf(stderr,"[textfile ...]\n\n");
  fprintf(stderr,"grappe match a set of patterns with limited and unlimited wildcards\n\n");
  fprintf(stderr,"summary of options:\n");
  fprintf(stderr,"-%c: to not ignore \"end of record\" characters (by default are ignored)\n",SKIP_EOR);
  fprintf(stderr,"-%c: before the pattern, useful when the pattern begins with a `-'\n",PATTERN);
  fprintf(stderr,"-%c: just count the occurence of successful matching in files\n",COUNTER_MODE);
#ifndef DNA
  fprintf(stderr,"-%c: case insensitive\n",CASE_SENSITIVE);
#endif
  fprintf(stderr,"-%c: line mode\n",LINE_MODE);
  fprintf(stderr,"-%c: to redefine what characters should be considered as end of records\n",DEFINE_SEPARATORS);
  fprintf(stderr,"-%c: before the name of the patternfile\n",PATTERN_FILE);
  fprintf(stderr,"Note that -%c and -%c are mutual exclusive options.",PATTERN,PATTERN_FILE); 
  fprintf(stderr,"\n");
  exit(WRONG_USAGE);
}

static int parse_options(int argc,char *argv[])
{
  char c;

  separators=NULL;

  while(((--argc)>0) && (*((++argv)[0])=='-')) // lazy evaluation !! (be careful, it can be a source of bugs...)
    {
      c=*(argv[0]+1);
      //fprintf(stderr,"%c\n",c);
      switch(c)
	{
	case PATTERN:
	  argc--;
	  if(argc <= 0)
	    {
	      fprintf(stderr,"%s: pattern must immediately follow option -%c.\n",progname,PATTERN);
	      usage();
	    }
	  if(pattern_filename != NULL)
	    {
	      fprintf(stderr,"%s: only one pattern should be specified.\n",progname);
	      usage();
	    }
	  pattern=(++argv)[0];
	  break;
	case PATTERN_FILE:
	  argc--;
	  if(argc <= 0)
	    {
	      fprintf(stderr,"%s: pattern filename must immediately follow option -%c.\n",progname,PATTERN_FILE);
	      usage();
	    }
	  if(pattern != NULL)
	    {
	      fprintf(stderr,"%s: only one set of patterns should be specified.\n",progname);
	      usage();
	    }
	  pattern_filename=(++argv)[0];
	  break;
	case SKIP_EOR:
	  skip_eor=(0==1);
	  break;
	case COUNTER_MODE:
	  counter_mode=(0==0);
	  break;
#ifndef DNA
	case CASE_SENSITIVE:
	  case_sensitive=(0==1);
	  break;
#endif
	case DEFINE_SEPARATORS:
	  argc--;
	  if(argc <= 0)
	    {
	      fprintf(stderr,"%s: the string containing the separators should immediately follow -%c.\n",progname,DEFINE_SEPARATORS);
	      usage();
	    }
	  if(separators != NULL)
	    {
	      fprintf(stderr,"%s: separators can't be specified more than one time.\n",progname);
	      usage();
	    }
	  separators=(++argv)[0];
	  break;
	case LINE_MODE:
	  line_mode=(0==0);
	  break;
	default:
	  fprintf(stderr,"%s: unrecognized option -%c.\n",progname,c);
	  usage();
	}
    }
  return argc;
}

// parse_cmdline parses the command line and returns the number of argument that are not options
// it also loads the patterns into the corresponding array
int parse_cmdline(int argc,char *argv[],int *nb_patterns)
{
  int tmp;

  progname=argv[0];

  if(argc < 2)
  {
    usage();
  }

  tmp=parse_options(argc,argv)+1;
  argv=argv+(argc-tmp);
  argc=tmp;
    
  if((pattern_filename==NULL) && (pattern==NULL))
  {
    --argc;
    if(argc <= 0)
    {
      fprintf(stderr,"A pattern or a patternfile must be specified.\n");
      usage();
    }
      
    pattern=(++argv)[0];
  }

#if 0==1
  if(pattern != NULL)
  {
    fprintf(stdout,"pattern: %s\n",pattern);
  }
  if(pattern_filename != NULL)
  {
    fprintf(stdout,"pattern filename: %s\n",pattern_filename);
  }
#endif

#ifndef DNA
  init_alphabet();
#else
  init_iupac_code();
#endif

  *nb_patterns=parse_pattern(pattern,pattern_filename);

  return (--argc);

#if 0==1
  if(--argc <= 0)
  {
    return NULL;
  }
  else
  {
    return (++argv)[0];
  }

  return NULL;
#endif
}
