#include <stdlib.h>
#include <assert.h>
#include "alphabet.h"
#include "utils.h"
#include "pattern.h"
#include "parse_pattern.h"
#include "errors.h"
#include "defs.h"

extern pattern *p[];

static char *pattern_string=NULL;
static FILE *pattern_file=NULL;

static char next_char_file(void)
{
  char c;
  c=fgetc(pattern_file);
  if(c == EOF)
    {
      return '\0';
    }
  else
    {
      return c;
    }
}

static char next_char_string(void)
{
  return *(pattern_string++);
}

static int parse_patterns(char (*next_char)(void))
{
  char c;
  char *buffer;
  int index=0;
  static int size_buffer=DFLT_SIZE;
  int couple=(0==1);
  int nb_comma=0;
  int read_nb=0;
  int wildcard=(0==1);
  int limited_wildcard=(0==1);
  int question=(0==1);
  int bracket_open=(0==1);
  int nb_patterns=0;
#ifdef SEMI_BOUNDED
  int semi_limited=(0==1);
#endif
  int index_open_bracket=0;


  buffer=(char *)malloc(size_buffer*sizeof(char));
  assert(buffer != NULL);

  while((c=next_char()) != '\0')
    {
      switch(get_letter_type(c))
	{
	case PEOR:
	case EOR:
	case SKIP:
	  continue;
	default:
	  break;
	}

      if(index+2 >= size_buffer)
	{
	  size_buffer=2*size_buffer;
	  buffer=(char *)realloc(buffer,size_buffer*sizeof(char));
	  assert(buffer != NULL);
	}

      if(couple)
	{
	  switch(c)
	    {
	    case RIGHT_PAREN:
	      if((read_nb > 0) && (nb_comma == 1))
		{
		  couple=(0==1);
		  limited_wildcard=(0==0);
#ifdef SEMI_BOUNDED
		  semi_limited=(0==1);
#endif
		}
	      else
		{
		  fprintf(stderr,"pattern not well formed: illegal limited wildcard sequence\n");
		  exit(PARSE_ERROR);
		}
	      buffer[index++]=c;
	      break;
	    case COMMA:
	      nb_comma++;
	      if((nb_comma > 1) || (read_nb <= 0))
		{
		  fprintf(stderr,"pattern not well formed: illegal limited wildcard sequence\n");
		  exit(PARSE_ERROR);
		}
	      buffer[index++]=c;
	      read_nb=0;
	      break;
#ifdef SEMI_BOUNDED
	    case STAR:
	      if((nb_comma == 1) && (read_nb == 0))
		{
		  buffer[index++]=c;
		  semi_limited=(0==0);
		  read_nb++;
		}
	      else
		{
		  fprintf(stderr,"pattern not well formed: illegal semi-limited wildcard sequence\n");
		  exit(PARSE_ERROR);
		}
	      break;
#endif
	    case '0':
	    case '1':
	    case '2':
	    case '3':
	    case '4':
	    case '5':
	    case '6':
	    case '7':
	    case '8':
	    case '9':
#ifdef SEMI_BOUNDED
	      if(!semi_limited)
		{
#endif
		  buffer[index++]=c;
		  read_nb++;
#ifdef SEMI_BOUNDED
		}
	      else
		{
		  fprintf(stderr,"pattern not well formed: illegal semi-limited wildcard sequence\n");
		  exit(PARSE_ERROR);
		}
#endif
	      break;
	    default:
	      fprintf(stderr,"pattern not well formed: illegal limited wildcard sequence (illegal character)\n");
	      exit(PARSE_ERROR);
	    }
	}
      else
	{
	  switch(c)
	    {
	    case LEFT_PAREN:
	      if((index <= 0) || (wildcard) || (limited_wildcard) || (bracket_open))
		{
		  fprintf(stderr,"pattern not well formed: illegal limited wildcard sequence\n");
		  exit(PARSE_ERROR);
		}
	      buffer[index++]=c;
	      couple=(0==0);
	      nb_comma=0;
	      read_nb=0;
#ifdef SEMI_BOUNDED
	      semi_limited=(0==1);
#endif
	      break;
	    case RIGHT_PAREN:
	      fprintf(stderr,"pattern not well formed (unprotected parenthesis)\n");
	      exit(PARSE_ERROR);
	      break;
	    case LEFT_SQ_BRACKET:
	      if(bracket_open)
		{
		  fprintf(stderr,"pattern not well formed (unprotected bracket)\n");
		  exit(PARSE_ERROR);
		}
	      index_open_bracket=index;
	      buffer[index++]=c;
	      bracket_open=(0==0);
	      break;
	    case RIGHT_SQ_BRACKET:
	      if(!bracket_open)
		{
		  fprintf(stderr,"pattern not well formed (mismatched bracket)\n");
		}
	      buffer[index++]=c;
	      bracket_open=(0==1);
	      break;
	    case CIRCUMFLEX:
	      if((!bracket_open) || (index != (index_open_bracket+1)))
		{
		  fprintf(stderr,"pattern not well formed (unprotected circumflex)\n");
		  exit(PARSE_ERROR);
		}
	      buffer[index++]=c;
	      break;
	    case QUESTION:
	      if((index > 0) && (!wildcard) && (!limited_wildcard) && (!question))
		{
		  fprintf(stderr,"pattern not well formed (unprotected question tag)\n");
		  exit(PARSE_ERROR);
		}
	      buffer[index++]=c;
	      question=(0==0);
	      break;
	    case DONTCARE:
	      if((limited_wildcard) || (bracket_open) || (question))
		{
		  fprintf(stderr,"pattern not well formed\n");
		  exit(PARSE_ERROR);
		}
	      if((index > 0) && (!wildcard))
		{
		  buffer[index++]=c;
		}
	      wildcard=(0==0);
	      break;
	    case PIPE:
	      if((limited_wildcard) || (wildcard) || (bracket_open) || (question))
		{
		  fprintf(stderr,"pattern not well formed\n");
		  exit(PARSE_ERROR);
		}
	      buffer[index]='\0';
	      if(index > 0)
		{
		  if(nb_patterns >= MAXPAT)
		    {
		      fprintf(stderr,"too many patterns... recompile with a higher MAXPAT value\n");
		      exit(PARSE_ERROR);
		    }
		  p[nb_patterns++]=compute_pattern(buffer);
		}
	      index=0;
	      couple=(0==1);
	      wildcard=(0==1);
	      limited_wildcard=(0==1);
	      question=(0==1);
	      read_nb=0;
	      nb_comma=0;
	      bracket_open=(0==1);
	      break;
	    case BACKSLASH:
	      buffer[index++]=c;
	      c=next_char();
	      if(c == '\0')
		{
		  fprintf(stderr,"pattern not well formed: %c sequence not finished\n",BACKSLASH);
		  exit(PARSE_ERROR);
		}
	      // we act as if it was a usual character
	    default:
	      if((get_letter_type(c) != DIRECT) && (get_letter_type(c) != INDIRECT))
		{
		  fprintf(stderr,"illegal character in pattern\n");
		  exit(PARSE_ERROR);
		}
	      wildcard=(0==1);	     
	      limited_wildcard=(0==1);
	      question=(0==1);
	      buffer[index++]=c;
	    }
	}
    }
  if(couple)
    {
      fprintf(stderr,"pattern not well formed (limited keyword not finished)\n");
      exit(PARSE_ERROR);
    }

  if((limited_wildcard) || (wildcard) || (bracket_open) || (question))
    {
      fprintf(stderr,"pattern not well formed\n");
      exit(PARSE_ERROR);
    }

  buffer[index]='\0';
  if(index > 0)
    {
      // fprintf(stderr,"%s\n",buffer);
      if(nb_patterns >= MAXPAT)
	{
	  fprintf(stderr,"too many patterns... recompile with a higher MAXPAT value\n");
	  exit(PARSE_ERROR);
	}
      p[nb_patterns++]=compute_pattern(buffer);
    }

  free(buffer);

  return nb_patterns;
}

int parse_pattern(char *pattern,char *pattern_filename)
{
 // exactly one of these pointers (pattern,pattern_filename) is NULL
  int nb_patterns;

  pattern_file=NULL;
  pattern_string=NULL;

  if(pattern == NULL)
  {
    assert(pattern_filename != NULL);
    pattern_file=fopen(pattern_filename,"r");
    if(pattern_file == NULL)
    {
      fprintf(stderr,"error while opening pattern file %s\n",pattern_filename);
      exit(FILE_ERROR);
    }
    nb_patterns=parse_patterns(next_char_file);
    fclose(pattern_file);
  }
  else
  {
    pattern_string=pattern;
    nb_patterns=parse_patterns(next_char_string);
  }
  return nb_patterns;
}

