#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "lists.h"
#include "utils.h"
#include "fragment.h"
#include "dawg.h"
#include "pattern.h"

static pattern_fragment *new_pattern_fragment(const char *frag)
{
  pattern_fragment *res;

  res=(pattern_fragment *)malloc(sizeof(pattern_fragment));
  assert(res != NULL);

  res->occurs=(0==1);
  res->end_position=-1;
  res->fragment=compute_fragment(frag);

  return res;
}

void reset_pattern(pattern *pat,node *source)
{
  pattern_fragment_list l;
  pattern_fragment *pat_frag;

  pat->current_fragment=pat->fragments;

  l=pat->fragments;
  while(!is_empty_pattern_fragment_list(l))
    {
      pat_frag=((pattern_fragment *)pattern_fragment_list_car(l));
      pat_frag->occurs=(0==1);
      pat_frag->end_position=-1;
      reset_fragment(pat_frag->fragment,source);
      l=pattern_fragment_list_cdr(l);
    }
}

void print_pattern(pattern *pat)
{
  pattern_fragment_list l;

  l=pat->fragments;
  while(!is_empty_pattern_fragment_list(l))
    {
      print_fragment(((pattern_fragment *)pattern_fragment_list_car(l))->fragment);
      l=pattern_fragment_list_cdr(l);
      if(!is_empty_pattern_fragment_list(l))
	{
	  fprintf(stdout,"#");
	}
    }
}

void print_pattern_position(pattern *pat)
{
  pattern_fragment_list l;
  pattern_fragment *pat_frag;

  pat->current_fragment=pat->fragments;

  l=pat->fragments;
  while(!is_empty_pattern_fragment_list(l))
    {
      pat_frag=((pattern_fragment *)pattern_fragment_list_car(l));
      if(pat_frag->occurs)
	{
	  fprintf(stdout,"%d ",pat_frag->end_position);
	}
      else
	{
	  fprintf(stdout,"? ");
	}
      l=pattern_fragment_list_cdr(l);
    }
}

/***********************************************/
/* pattern_string is assumed to be well formed */
/* it must follow the rules:                   */
/* pattern = fragment | fragment#pattern       */
/***********************************************/
pattern *compute_pattern(const char *pattern_string)
{
  char c;
  char *p,*q,*copy;
  pattern *pat;

  pat=(pattern *)malloc(sizeof(pattern));
  assert(pat != NULL);

  pat->fragments=empty_pattern_fragment_list();

  copy=strdup(pattern_string);
  assert(copy != NULL);

  q=copy;
  p=q;
  while((c=*(q++)) != '\0')
    {
      switch(c)
      {
      case DONTCARE:
	*(--q)='\0';
	pat->fragments=pattern_fragment_list_append(pat->fragments,pattern_fragment_list_cons((struct pattern_fragment *)new_pattern_fragment(p),empty_pattern_fragment_list()));
	*(q++)=DONTCARE;
	p=q;
	break;
      case BACKSLASH:
	c=*(q++);
	assert(c != '\0');
      default:
	// nothing
      }
    }
  pat->fragments=pattern_fragment_list_append(pat->fragments,pattern_fragment_list_cons((struct pattern_fragment *)new_pattern_fragment(p),empty_pattern_fragment_list()));

  free(copy);

  pat->underloading=(0==1);
  pat->current_fragment=pat->fragments; // to the start of the pattern
  return pat;
}

void load_pattern(pattern *pat,int position,node *source)
{
  if(pat->underloading)
    {
      if(!is_empty_pattern_fragment_list(pat->current_fragment))
	{
	  pattern_fragment *pat_frag;
	  pat_frag=((pattern_fragment *)pattern_fragment_list_car(pat->current_fragment));
	  if(load_fragment(pat_frag->fragment,position,source,pat,pat_frag))
	    {
	      pat->underloading=(0==1);
	      pat->current_fragment=pattern_fragment_list_cdr(pat->current_fragment);
	    }
	}
    }
  clean_fragment(((pattern_fragment *)pattern_fragment_list_car(pat->fragments))->fragment,position);
}

int pattern_occurs(pattern *pat)
{
  pattern_fragment_list l;
  pattern_fragment *pat_frag;
  l=pat->fragments;
  while(!is_empty_pattern_fragment_list(l))
    {
      pat_frag=((pattern_fragment *)pattern_fragment_list_car(l));
      if(!pat_frag->occurs)
	{
	  return (0==1);
	}
      l=pattern_fragment_list_cdr(l);
    }
  return (0==0);
}
